import logging
import numpy as np

from datetime import datetime, timedelta
from collections import OrderedDict


logger = logging.getLogger(__name__)


class TrackedObject:
    def __init__(self, pos: np.ndarray):
        self.pos = pos
        self.last_seen = self._now()
        self.id = None
        self.tracking_ctx = None
        self.created = self._now()
        self.history = []

    def __hash__(self):
        return int(id(self) / 16)

    def __eq__(self, other):
        if isinstance(other, TrackedObject):
            return self.id == other.id
        return False

    def __str__(self):
        return "Tracked Object {{id = {}, pos = {}, last_seen = {}}}".format(self.id, self.pos, self.last_seen)

    def _now(self):
        return datetime.now()


class ObjectTracker:
    def __init__(self, deletion_threshold: int = 5):
        self._objects = OrderedDict()
        self._next = 0
        self._deletion_threshold = deletion_threshold

    def update(self, objects: [TrackedObject]):
        new_objs = self._track(objects)

        for obj in new_objs:
            self._register(obj)

        self._prune()

    def get_objects(self) -> [TrackedObject]:
        return list(self._objects.values())

    def _prune(self):
        now = self._now()
        to_prune = []
        for obj in self._objects.values():
            time_since_last_seen = (now - obj.last_seen) / timedelta(seconds=1)
            time_since_created = (now - obj.created) / timedelta(seconds=1)
            if time_since_last_seen > self._deletion_threshold:
                # delete if it has been too long since we've seen the object
                logger.debug("pruning {} because it hasn't been seen in {} seconds".format(obj.id,
                                                                                           time_since_last_seen))
                to_prune.append(obj)
            elif time_since_created > self._deletion_threshold and time_since_last_seen / time_since_created > .6:
                # delete if we haven't seen most its life
                logger.debug("pruning {} because it hasn't been seen for the majority of its life".format(obj.id))
                to_prune.append(obj)

        for obj in to_prune:
            self._objects.pop(obj.id)
            logger.debug("pruned: {}".format(obj))

    def _register(self, obj: TrackedObject):
        obj.id = self._get_next_object_id()
        self._objects[obj.id] = obj
        logger.debug("registered new object: {}".format(obj))

    def _track(self, objects: [TrackedObject]) -> [TrackedObject]:
        return objects

    def _get_next_object_id(self) -> int:
        id = self._next
        self._next += 1
        return id

    def _now(self) -> datetime:
        return datetime.now()



import circum.utils.math

import numpy as np


def test_rotate_x():
    positions = np.array([[0, 0, 0],
                          [1, 0, 1],
                          [0, 1, 1],
                          [1, 1, 1]])

    sensor_pose = np.array([[0],
                            [0],
                            [0],
                            [90],
                            [0],
                            [0]])

    expected = np.array([[0,  0,  0],
                         [0, -1, -1],
                         [1,  0,  1],
                         [1,  1,  1]])

    transformed = circum.utils.math.transform_positions(positions, sensor_pose)

    assert np.allclose(expected, transformed)


def test_rotate_y():
    positions = np.array([[1, 0, 1],
                          [0, 0, 0],
                          [0, 1, 1],
                          [1, 1, 1]])

    sensor_pose = np.array([[0],
                            [0],
                            [0],
                            [0],
                            [90],
                            [0]])

    expected = np.array([[0,  1,  1],
                         [0,  0,  0],
                         [-1, 0, -1],
                         [1,  1,  1]])

    transformed = circum.utils.math.transform_positions(positions, sensor_pose)

    assert np.allclose(expected, transformed)


def test_rotate_z():
    positions = np.array([[1, 0, 1],
                          [0, 1, 1],
                          [0, 0, 0],
                          [1, 1, 1]])

    sensor_pose = np.array([[0],
                            [0],
                            [0],
                            [0],
                            [0],
                            [90]])

    expected = np.array([[0, -1, -1],
                         [1,  0,  1],
                         [0,  0,  0],
                         [1,  1,  1]])

    transformed = circum.utils.math.transform_positions(positions, sensor_pose)

    assert np.allclose(expected, transformed)


def test_rotate_translate():
    positions = np.array([[1, 0, 0, 1],
                          [0, 1, 0, 1],
                          [0, 0, 1, 1],
                          [1, 1, 1, 1]])

    sensor_pose = np.array([[10],
                            [20],
                            [30],
                            [45],
                            [90],
                            [135]])

    expected = np.array([[10,  9, 10,  9],
                         [20, 20, 21, 21],
                         [29, 30, 30, 29],
                         [1,   1,  1,  1]])

    transformed = circum.utils.math.transform_positions(positions, sensor_pose)

    assert np.allclose(expected, transformed)



import click
import circum.endpoint
import copy
import logging
import numpy as np
import random
import time

from threading import Semaphore, Thread


logger = logging.getLogger(__name__)
tracking_semaphore = None
tracking_info = {"objects": []}
vector_info = []
updated = False


def _update_thread(update_interval: float, num_objects: int):
    global tracking_info
    global vector_info
    global updated

    while True:
        tracking_semaphore.acquire()
        if len(vector_info) == 0:
            if num_objects == 0:
                num_objects = 1 + int(random.random() * 4)
            for i in range(num_objects):
                # position
                x = random.random() * 10
                y = random.random() * 10
                z = random.random() * 3
                pos = np.array([x, y, z])

                # velocity
                dx = random.random()
                dy = random.random()
                vel = np.array([dx, dy, 0])
                mag = random.random() * 2
                vel = mag * vel / np.linalg.norm(vel)

                vector_info.append({"pos": pos, "vel": vel})
        else:
            # update
            for obj in vector_info:
                for i in range(len(obj["vel"])):
                    if obj["pos"][i] > 10 or obj["pos"][i] < 0:
                        obj["vel"][i] = -obj["vel"][i]
                obj["pos"] += obj["vel"]

        # track
        tracking_info["objects"] = [{"x": obj["pos"][0], "y": obj["pos"][1], "z": obj["pos"][2]} for obj in vector_info]
        updated = True

        tracking_semaphore.release()
        time.sleep(update_interval)


def run_simulator(simulator_args: {}) -> {}:
    global updated
    ret = None
    tracking_semaphore.acquire()
    if updated:
        ret = copy.deepcopy(tracking_info)
        updated = False
    tracking_semaphore.release()
    return ret


@click.command()
@click.option('--update_interval',
              required=False,
              type=float,
              default=.5,
              help='Rate to send updates.')
@click.option('--num_objects',
              required=False,
              type=int,
              default=4,
              help='Number of objects to simulate')
@click.pass_context
def simulator(ctx, update_interval: float, num_objects: int):
    global tracking_semaphore
    tracking_semaphore = Semaphore()
    logger.debug("simulating {} objects at {} Hz".format(num_objects, 1/update_interval))
    tracker_thread = Thread(target=_update_thread, args=[update_interval, num_objects])
    tracker_thread.daemon = True
    tracker_thread.start()
    circum.endpoint.start_endpoint(ctx, "simulator", run_simulator)



from circum.utils.state.simple_tracker import SimpleTracker
from circum.utils.state.tracking import TrackedObject
import copy
import numpy as np
import mock
import datetime


@mock.patch("circum.utils.state.tracking.ObjectTracker._now")
@mock.patch("circum.utils.state.tracking.TrackedObject._now")
def test_simple_update(object_now, tracker_now):
    now = datetime.datetime.now()

    object_now.side_effect = [
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
    ]

    tracker_now.side_effect = [
        now + datetime.timedelta(seconds=1),  # _track
        now + datetime.timedelta(seconds=1),  # _prune
    ]

    objects = [
        TrackedObject(np.array([0, 0, 0])),
        TrackedObject(np.array([1, 0, 0])),
        TrackedObject(np.array([0, 1, 0])),
        TrackedObject(np.array([1, 1, 0]))
    ]

    tracker = SimpleTracker(deletion_threshold=5)

    tracker.update(copy.deepcopy(objects))

    tracked_objects = tracker.get_objects()

    assert len(tracked_objects) == len(objects)

    for i in range(len(objects)):
        assert any(np.array_equal(obj.pos, objects[i].pos) for obj in tracked_objects)


@mock.patch("circum.utils.state.tracking.ObjectTracker._now")
@mock.patch("circum.utils.state.tracking.TrackedObject._now")
def test_simple_update_stable_association(object_now, tracker_now):
    now = datetime.datetime.now()

    object_now.side_effect = [
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
    ]

    tracker_now.side_effect = [
        now + datetime.timedelta(seconds=1),  # _track
        now + datetime.timedelta(seconds=1),  # _prune
        now + datetime.timedelta(seconds=1),  # _track
        now + datetime.timedelta(seconds=1),  # _prune
    ]

    objects = [
        TrackedObject(np.array([0, 0, 0])),
        TrackedObject(np.array([1, 0, 0])),
        TrackedObject(np.array([0, 1, 0])),
        TrackedObject(np.array([1, 1, 0]))
    ]

    tracker = SimpleTracker(deletion_threshold=5)

    tracker.update(copy.deepcopy(objects))

    tracked_objects1 = copy.deepcopy(tracker.get_objects())

    tracker.update(copy.deepcopy(objects))

    tracked_objects2 = copy.deepcopy(tracker.get_objects())

    assert len(tracked_objects1) == len(tracked_objects2)

    obj_map = {obj.id: obj for obj in tracked_objects1}

    for i in range(len(objects)):
        assert np.array_equal(tracked_objects2[i].pos, obj_map[tracked_objects2[i].id].pos)


@mock.patch("circum.utils.state.tracking.ObjectTracker._now")
@mock.patch("circum.utils.state.tracking.TrackedObject._now")
def test_simple_update_10_pct_delta(object_now, tracker_now):
    now = datetime.datetime.now()

    object_now.side_effect = [
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
    ]

    tracker_now.side_effect = [
        now + datetime.timedelta(seconds=1),  # _track
        now + datetime.timedelta(seconds=1),  # _prune
        now + datetime.timedelta(seconds=1),  # _track
        now + datetime.timedelta(seconds=1),  # _prune
    ]

    objects1 = [
        TrackedObject(np.array([0, 0, 0])),
        TrackedObject(np.array([1, 0, 0])),
        TrackedObject(np.array([0, 1, 0])),
        TrackedObject(np.array([1, 1, 0]))
    ]

    objects2 = [
        TrackedObject(np.array([0.1, 0, 0])),
        TrackedObject(np.array([0.9, 0, 0])),
        TrackedObject(np.array([0.1, 1, 0])),
        TrackedObject(np.array([0.9, 1, 0]))
    ]

    tracker = SimpleTracker(deletion_threshold=5)

    tracker.update(copy.deepcopy(objects1))

    tracked_objects1 = copy.deepcopy(tracker.get_objects())

    tracker.update(copy.deepcopy(objects2))

    tracked_objects2 = copy.deepcopy(tracker.get_objects())

    assert len(tracked_objects1) == len(tracked_objects2)

    obj_map = {obj.id: obj for obj in tracked_objects2}

    for i in range(len(objects1)):
        assert np.linalg.norm(tracked_objects1[i].pos - obj_map[tracked_objects1[i].id].pos) < .2


@mock.patch("circum.utils.state.tracking.ObjectTracker._now")
@mock.patch("circum.utils.state.tracking.TrackedObject._now")
def test_simple_update_permanence(object_now, tracker_now):
    now = datetime.datetime.now()

    object_now.side_effect = [
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
        now,  # create
    ]

    tracker_now.side_effect = [
        now + datetime.timedelta(seconds=1),  # _track
        now + datetime.timedelta(seconds=1),  # _prune
        now + datetime.timedelta(seconds=1),  # _track
        now + datetime.timedelta(seconds=1),  # _prune
    ]

    objects = [
        TrackedObject(np.array([0, 0, 0])),
        TrackedObject(np.array([1, 0, 0])),
        TrackedObject(np.array([0, 1, 0])),
        TrackedObject(np.array([1, 1, 0]))
    ]

    tracker = SimpleTracker(deletion_threshold=5)

    tracker.update(copy.deepcopy(objects))

    tracked_objects1 = copy.deepcopy(tracker.get_objects())

    tracker.update([])

    tracked_objects2 = copy.deepcopy(tracker.get_objects())

    assert len(tracked_objects1) == len(tracked_objects2)

    obj_map = {obj.id: obj for obj in tracked_objects1}

    for i in range(len(objects)):
        assert np.array_equal(tracked_objects2[i].pos, obj_map[tracked_objects2[i].id].pos)



from circum.utils.state.kalman.kalmanfilter import KalmanFilter
import numpy as np
import datetime


class EKF:
    """
    A class that gets sensor measurements from class DataPoint
    and predicts the next state of the system using an extended Kalman filter algorithm
    The state variables we are considering in this system are the position and velocity
    in x and y cartesian coordinates, in essence there are 4 variables we are tracking.
    """

    def __init__(self, d: {}):
        self.initialized = False
        self.timestamp = None
        self.n = d['number_of_states']
        self.P = d['initial_process_matrix']
        self.F = d['inital_state_transition_matrix']
        self.Q = d['initial_noise_matrix']
        self.R = d['covariance_matrix']
        self.H = d['transition_matrix']
        self.a = d['acceleration_noise']
        self.kalmanFilter = KalmanFilter(self.n)

    def updateQ(self, dt: int):
        dt2 = dt * dt
        dt3 = dt * dt2
        dt4 = dt * dt3

        ula = [dt4 * a_k / 4 for a_k in self.a]
        ura = [dt3 * a_k / 2 for a_k in self.a]
        lla = [dt3 * a_k / 2 for a_k in self.a]
        lra = [dt2 * a_k for a_k in self.a]

        half_n = int(self.n/2)

        ul = np.zeros((half_n, half_n))
        ur = np.zeros((half_n, half_n))
        ll = np.zeros((half_n, half_n))
        lr = np.zeros((half_n, half_n))

        np.fill_diagonal(ul, ula)
        np.fill_diagonal(ur, ura)
        np.fill_diagonal(ll, lla)
        np.fill_diagonal(lr, lra)

        Q = np.concatenate((np.concatenate((ul, ur), axis=1), np.concatenate((ll, lr), axis=1)))

        self.kalmanFilter.set_Q(Q)

    def predict(self, timestamp: datetime.datetime):
        dt = (timestamp - self.timestamp) / datetime.timedelta(seconds=1)
        self.timestamp = timestamp

        self.kalmanFilter.update_F(dt)
        self.updateQ(dt)
        self.kalmanFilter.predict()

    def update(self, data: np.ndarray, timestamp: datetime.datetime):
        z = data.reshape(data.shape[0], 1)
        x = self.kalmanFilter.get_x()

        Hx = self.H @ x

        self.kalmanFilter.update(z, self.H, Hx, self.R)
        self.timestamp = timestamp

    def predict_update(self, data: np.ndarray, timestamp: datetime.datetime):
        self.predict(timestamp)
        self.update(data, timestamp)

    def start(self, data: np.ndarray, timestamp: datetime.datetime):
        self.timestamp = timestamp
        x = np.concatenate((data, np.array([0, 0, 0])))
        x = x.reshape(x.shape[0], 1)
        self.kalmanFilter.start(x, self.P, self.F, self.Q)
        self.initialized = True

    def process(self, data: (float, float), timestamp: int):
        if self.initialized:
            self.update(data, timestamp)
        else:
            self.start(data, timestamp)

    def get(self):
        return self.kalmanFilter.get_x().T



from circum.utils.state.tracking import TrackedObject, ObjectTracker
from datetime import datetime, timedelta
import numpy as np
from collections import OrderedDict
import copy
import mock


def test_TrackedObject_creation():
    time_thresh = timedelta(milliseconds=100)
    to = TrackedObject(np.array([0, 1, 2]))
    dt = datetime.now()
    assert np.array_equal(to.pos, np.array([0, 1, 2]))
    assert dt - to.last_seen < time_thresh
    assert to.id is None
    assert to.tracking_ctx is None
    assert dt - to.created < time_thresh
    assert to.history == []


def test_TrackedObject_equal():
    to1 = TrackedObject(np.array([0, 1, 2]))
    to2 = TrackedObject(np.array([0, 1, 2]))

    to1.id = 1
    to2.id = 1

    assert to1 == to2


def test_TrackedObject_notequal_different_ids():
    to1 = TrackedObject(np.array([0, 1, 2]))
    to2 = TrackedObject(np.array([0, 1, 2]))

    to1.id = 1
    to2.id = 2

    assert to1 != to2


def test_TrackedObject_notequal_different_types():
    to1 = TrackedObject(np.array([0, 1, 2]))
    to2 = object()

    assert to1 != to2


def test_ObjectTracker_creation():
    ot = ObjectTracker(5)
    assert isinstance(ot._objects, OrderedDict)
    assert len(ot._objects) == 0
    assert ot._next == 0
    assert ot._deletion_threshold == 5


def test_ObjectTracker_update():
    ot = ObjectTracker(5)
    tos = [TrackedObject(np.array([0, 1, 2]))]
    ot.update(tos)
    assert len(ot._objects) == 1
    assert ot._objects[0].id == 0


def test_ObjectTracker_update_multiple():
    ot = ObjectTracker(5)
    tos = [TrackedObject(np.array([0, 1, 2])), TrackedObject(np.array([0, 1, 2]))]
    ot.update(tos)
    assert len(ot._objects) == 2
    assert ot._objects[0].id == 0
    assert ot._objects[1].id == 1


def test_ObjectTracker_prune_too_old():
    dt = datetime.now()
    ot = ObjectTracker(5)
    tos = [TrackedObject(np.array([0, 1, 2])), TrackedObject(np.array([0, 1, 2]))]
    ot.update(tos)

    assert len(ot._objects) == 2

    ot._objects[1].created = dt - timedelta(seconds=100)
    ot._objects[1].last_seen = dt - timedelta(seconds=6)

    ot.update([])

    assert len(ot._objects) == 1


@mock.patch("circum.utils.state.tracking.ObjectTracker._now")
@mock.patch("circum.utils.state.tracking.TrackedObject._now")
def test_ObjectTracker_prune_too_invisible(object_now, tracker_now):
    now = datetime.now()

    object_now.side_effect = [
        now,  # create
        now,  # create
        now,  # create
        now,  # create
    ]

    tracker_now.side_effect = [
        # update 1
        now,  # _prune
        # update 2
        now + timedelta(seconds=10),  # _prune
    ]

    ot = ObjectTracker(5)
    tos = [TrackedObject(np.array([0, 1, 2])), TrackedObject(np.array([0, 1, 2]))]

    ot.update(copy.deepcopy(tos))
    assert len(ot._objects) == 2

    ot._objects[0].last_seen = now + timedelta(seconds=9)
    ot._objects[1].last_seen = now + timedelta(seconds=3.9)

    ot.update([])
    assert len(ot._objects) == 1



import math

import numpy as np
from math import cos
from math import sin


def transform_positions(positions: np.ndarray, sensor_pose: [float]):
    translation_matrix = np.identity(4)
    translation_matrix[0][3] = sensor_pose[0]
    translation_matrix[1][3] = sensor_pose[1]
    translation_matrix[2][3] = sensor_pose[2]

    theta_x = math.radians(sensor_pose[3])
    theta_y = math.radians(sensor_pose[4])
    theta_z = math.radians(sensor_pose[5])

    rx = np.array([[1,     0,              0,       0],
                   [0, cos(theta_x), -sin(theta_x), 0],
                   [0, sin(theta_x),  cos(theta_x), 0],
                   [0,     0,              0,       1]])

    ry = np.array([[cos(theta_y),  0, sin(theta_y), 0],
                   [0,             1,       0,      0],
                   [-sin(theta_y), 0, cos(theta_y), 0],
                   [0,             0,       0,      1]])

    rz = np.array([[cos(theta_z), -sin(theta_z), 0, 0],
                   [sin(theta_z),  cos(theta_z), 0, 0],
                   [0,                   0,      1, 0],
                   [0,                   0,      0, 1]])

    # first rotate around the sensor's angles, these use the sensor location as the origin
    # then, translate, setting the positions' origin to the system's origin
    return translation_matrix @ (rz @ (ry @ (rx @ positions)))



import logging
import numpy as np
import scipy.spatial.distance as dist

from circum.utils.state.tracking import TrackedObject, ObjectTracker
from circum.utils.state.kalman.ekf import EKF
from munkres import Munkres


logger = logging.getLogger(__name__)


class KalmanContext:
    def __init__(self):
        num_states = 6
        half_num_states = int(num_states/2)
        R = np.zeros([half_num_states, half_num_states])
        np.fill_diagonal(R, 0.01)

        H = np.zeros((half_num_states, num_states))
        np.fill_diagonal(H, 1)

        P_vals = [1 for i in range(half_num_states)] + [1000 for i in range(half_num_states)]
        P = np.zeros([num_states, num_states])
        np.fill_diagonal(P, P_vals)

        Q = np.zeros([num_states, num_states])
        F = np.eye(num_states)

        d = {
            'number_of_states': 6,
            'initial_process_matrix': P,
            'covariance_matrix': R,
            'transition_matrix': H,
            'inital_state_transition_matrix': F,
            'initial_noise_matrix': Q,
            'acceleration_noise': (5, 5, 5)
        }

        self.kf = EKF(d)


class KalmanTracker(ObjectTracker):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _associate(self,
                   detected: [TrackedObject],
                   threshold: float = 10) -> ([(TrackedObject, TrackedObject)], [TrackedObject], [TrackedObject]):
        tracked_objects = self.get_objects()
        object_positions = [obj.pos for obj in tracked_objects]
        new_positions = [obj.pos for obj in detected]
        if len(tracked_objects) == 0:
            return [], detected, []
        if len(detected) == 0:
            return [], [], tracked_objects
        distances = dist.cdist(np.asarray(new_positions), np.asarray(object_positions)).tolist()

        m = Munkres()
        indexes = m.compute(distances)

        associations = []
        associated_detections = [detected[detected_index] for detected_index, _ in indexes]
        associated_tracked = [tracked_objects[tracked_index] for _, tracked_index in indexes]
        unassociated_detections = set(detected) - set(associated_detections)
        unassociated_tracked = set(tracked_objects) - set(associated_tracked)

        # unassociate anything that jumped too far
        for row, column in indexes:
            if distances[row][column] < threshold:
                associations.append((tracked_objects[column], detected[row]))
            else:
                unassociated_detections.add(detected[row])
                unassociated_tracked.add(tracked_objects[column])

        return associations, unassociated_detections, unassociated_tracked

    def _predict(self):
        # update all of the currently tracked objects with their predictions
        tracked_objects = self.get_objects()
        now = self._now()
        for tracked_object in tracked_objects:
            tracked_object.tracking_ctx.kf.predict(now)
            tracked_object.pos =\
                np.array(tracked_object.tracking_ctx.kf.get()[0, 0:int(tracked_object.tracking_ctx.kf.n / 2)])

    def _track(self, objects: [TrackedObject]) -> [TrackedObject]:
        # predict
        self._predict()

        associations, unassociated_detections, unassociated_tracked = self._associate(objects)

        now = self._now()

        # add context to new objects
        for detection in unassociated_detections:
            detection.tracking_ctx = KalmanContext()
            detection.tracking_ctx.kf.start(detection.pos, now)

        # predict/update
        for tracked, detection in associations:
            tracked.tracking_ctx.kf.update(detection.pos, now)
            tracked.last_seen = now
            tracked.pos = np.array(tracked.tracking_ctx.kf.get()[0, 0:int(tracked.tracking_ctx.kf.n / 2)])

        return unassociated_detections



import logging
import numpy as np
import scipy.spatial.distance as dist

from circum.utils.state.tracking import TrackedObject, ObjectTracker


logger = logging.getLogger(__name__)


class SimpleTracker(ObjectTracker):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _track(self, objects: [TrackedObject]) -> [TrackedObject]:
        now = self._now()
        tracked_objects = self.get_objects()

        if len(tracked_objects) == 0:
            return objects
        elif len(objects) == 0:
            return []
        else:
            object_ids = [obj.id for obj in tracked_objects]
            object_positions = [obj.pos for obj in tracked_objects]
            new_positions = [obj.pos for obj in objects]

            distances = dist.cdist(np.asarray(object_positions), np.asarray(new_positions))

            rows = distances.min(axis=1).argsort()
            cols = distances.argmin(axis=1)[rows]

            used_rows = set()
            used_cols = set()

            for (row, col) in zip(rows, cols):
                if row in used_rows or col in used_cols:
                    continue

                id_ = object_ids[row]
                self._objects[id_].pos = new_positions[col]
                self._objects[id_].history.append(new_positions[col])
                self._objects[id_].last_seen = now

                # indicate that we have examined each of the row and
                # column indexes, respectively
                used_rows.add(row)
                used_cols.add(col)

            if distances.shape[0] < distances.shape[1]:
                unused_cols = set(range(0, distances.shape[1])).difference(used_cols)
                return [objects[col] for col in unused_cols]
            else:
                return []



import logging
import socket
import sys
import ifaddr

from ipaddress import IPv4Address, AddressValueError
from zeroconf import Zeroconf, ServiceInfo, get_all_addresses


logger = logging.getLogger(__name__)


def _advertise_server(name: str, type: str, ips: [str], port: int, properties=b"") -> (Zeroconf, [ServiceInfo]):
    desc = properties

    service_type = "_{}._sub._circum._tcp.local.".format(type)
    service_name = "{}.{}".format(name, service_type)

    zeroconf = Zeroconf()
    infos = []

    logger.debug("registering {} at {}:{}".format(service_name, ips, port))
    info = ServiceInfo(service_type,
                       service_name,
                       [socket.inet_aton(ip) for ip in ips], port, 0, 0,
                       desc)
    zeroconf.register_service(info)
    infos.append(info)

    print("Registered service, press Ctrl-C to exit...")
    return zeroconf, infos


def _open_server(ips: [str], port: int) -> [socket.socket]:
    sockets = []
    socket_ips = []
    for ip in ips:
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.bind((ip, port))
            s.listen()
            sockets.append(s)
            socket_ips.append(ip)
        except Exception:
            # not all ips on all interfaces will be bindable
            logger.warning("Unable to bind to address: {}".format(ip))
    return sockets, socket_ips


def _get_interface_ip(interface: str) -> [str]:
    if interface is None:
        return get_all_addresses()

    # check if the interface was passed as an IP
    try:
        IPv4Address(interface)
        return [interface]
    except AddressValueError:
        pass

    # pyzeroconf only supports ipv4 so limit to that
    ipv4s = list(
        set(
            addr.ip
            for iface in ifaddr.get_adapters()
            for addr in iface.ips
            # Host only netmask 255.255.255.255
            if addr.is_IPv4 and iface.nice_name == interface and addr.network_prefix != 32
        )
    )

    if len(ipv4s) == 0:
        raise Exception("unable to find an ipv4 address for interface")

    return ipv4s


def _set_keepalive(sock: socket.socket, interval=1, retries=5):
    platform = sys.platform
    if platform.startswith('linux'):
        # linux specific keepalive
        logger.debug("setting linux keepalive options")
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, interval)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, interval)
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_KEEPCNT, retries)
    elif platform.startswith('win32'):
        # windows specific keepalive
        logger.debug("setting windows keepalive options")
        interval_ms = interval * 1000
        sock.ioctl(socket.SIO_KEEPALIVE_VALS, (1, interval_ms, interval_ms))
        # windows retry count is configured via registry value (max data retransmission attempts) and defaults to 5
    elif platform.startswith('darwin'):
        # osx specific keepalive
        logger.debug("setting OS X keepalive options")
        TCP_KEEPALIVE = 0x10
        TCP_KEEPINTVL = 0x101
        TCP_KEEPCNT = 0x102
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        sock.setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, interval)
        sock.setsockopt(socket.IPPROTO_TCP, TCP_KEEPINTVL, interval)
        sock.setsockopt(socket.IPPROTO_TCP, TCP_KEEPCNT, retries)
    # we'll leave the defaults on other os's


class ServiceListener:
    def __init__(self, services: [str]):
        self.sockets = {}
        self.services = services

    def remove_service(self, zeroconf, type_, name):
        logger.debug("Service {} removed".format(name))
        self.sockets.pop(name).close()

    def add_service(self, zeroconf, type_, name):
        info = zeroconf.get_service_info(type_, name)
        if len(self.services) > 0:
            if name not in self.services:
                logger.debug("Name doesn't match skipping {}, service info: {}".format(name, info))
                return
        logger.debug("Service {} added, service info: {}".format(name, info))
        if name not in self.sockets.keys():
            addresses = [str(IPv4Address(address)) for address in info.addresses]
            for address in addresses:
                try:
                    logger.debug("attempting to connects to ({}, {})".format(address, info.port))
                    service_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                    service_socket.connect((address, info.port))
                    _set_keepalive(service_socket)
                    self.sockets[name] = service_socket
                    logger.debug("connected")
                    break
                except Exception:
                    logger.warn("unable to create socket", exc_info=True)

    def remove(self, service_socket: socket.socket):
        for k, s in self.sockets.iteritems():
            if s == service_socket:
                self.sockets.pop(k).close()
                return

    def get_sockets(self):
        return list(self.sockets.values())



#!/bin/python3

import bson
import click
import logging
import select
import socket
import struct

import numpy as np

from circum.utils.network import _advertise_server, _open_server, _set_keepalive, _get_interface_ip, ServiceListener
from circum.utils.state.tracking import TrackedObject
from circum.utils.state.kalman_tracker import KalmanTracker as Tracker
from zeroconf import ServiceBrowser, Zeroconf
from threading import Semaphore


logger = logging.getLogger(__name__)
size_fmt = "!i"
size_data_len = struct.calcsize(size_fmt)
tracking_state = Tracker()


def _update(update: {}, clients: [socket.socket]):
    people = [TrackedObject(np.asarray([person["x"], person["y"], person["z"]])) for person in update]
    tracking_state.update(people)
    tracked = tracking_state.get_objects()
    update_dict = {
        "objects": [{"x": person.pos[0], "y": person.pos[1], "z": person.pos[2], "id": person.id} for person in tracked]
    }
    bson_data = bson.dumps(update_dict)
    length = len(bson_data)
    size_data = struct.pack(size_fmt, length)
    data = size_data + bson_data

    excepted = []

    for client in clients:
        try:
            client.sendall(data)
        except Exception:
            excepted.append(client)

    for client in excepted:
        clients.remove(client)


def _run_service(server_sockets: [socket.socket], listener: ServiceListener):
    semaphore = Semaphore()
    clients = []

    last_update = {}

    while True:
        endpoint_sockets = listener.get_sockets()

        removed_endpoints = set(last_update.keys()) - set(endpoint_sockets)
        for removed_endpoint in removed_endpoints:
            last_update.pop(removed_endpoint)

        # service the sockets
        ready, _, excepted = select.select(server_sockets + endpoint_sockets, [], [], 1)
        for ready_socket in ready:
            if ready_socket in server_sockets:
                conn, _ = ready_socket.accept()
                _set_keepalive(conn)
                semaphore.acquire()
                clients.append(conn)
                semaphore.release()
            elif ready_socket in endpoint_sockets:
                try:
                    size_data = ready_socket.recv(size_data_len)
                    size = struct.unpack(size_fmt, size_data)[0]
                    data = ready_socket.recv(size)
                    update_data = bson.loads(data)
                    last_update[ready_socket] = update_data["objects"]
                    _update([person for update in last_update.values() for person in update], clients)
                except OSError:
                    if ready_socket not in excepted:
                        excepted.append(ready_socket)
        for excepted_socket in excepted:
            if excepted_socket not in server_sockets:
                listener.remove(excepted_socket)


def _start_service(name: str, interface: str, port: int, listener: ServiceListener):
    ips = _get_interface_ip(interface)

    logger.debug("opening server on ({},{})".format(ips, port))
    server_sockets, ips = _open_server(ips, port)

    zeroconf, infos = _advertise_server(name, "service", ips, port)

    try:
        _run_service(server_sockets, listener)
    except Exception:
        logging.error("Exception while running server", exc_info=True)
    finally:
        for info in infos:
            zeroconf.unregister_service(info)
        zeroconf.close()


@click.command()
@click.option('--name',
              '-n',
              required=True,
              help='The service name')
@click.option('--interface',
              '-i',
              required=False,
              default=None,
              help='The interface to bind to.')
@click.option('--port',
              '-p',
              required=False,
              default=8300,
              type=int,
              help='The port to bind to.')
@click.option('--endpoint',
              '-e',
              multiple=True,
              type=str,
              help='Names of endpoints to connect to. Can be specified multiple times. ' +
                   'If no endpoints are specified, all available endpoints will be used.')
def cli(name: str, interface: str, port: int, endpoint: [str]):
    global logger
    logging.basicConfig(level="DEBUG")
    logger = logging.getLogger("circum_service")
    zeroconf = Zeroconf()
    endpoint_type = "_endpoint._sub._circum._tcp.local."
    listener = ServiceListener([name + "." + endpoint_type for name in endpoint])
    browser = ServiceBrowser(zeroconf, endpoint_type, listener)  # noqa
    try:
        _start_service(name, interface, port, listener)
    finally:
        zeroconf.close()


if __name__ == "__main__":
    cli()



#!/bin/python3

import bson
import click
import logging
import numpy as np
import pkg_resources
import socket
import struct
import uuid

from select import select
from threading import Semaphore, Thread
from typing import Callable
from circum.utils.network import _advertise_server, _open_server, _get_interface_ip, _set_keepalive
from circum.utils.math import transform_positions


logger = logging.getLogger(__name__)
circum_sensors = []
circum_pose_providers = {}


# register sensors
pose_providers = {entry_point.name: entry_point.load() for
                  entry_point in
                  pkg_resources.iter_entry_points('circum.pose_providers')}


def _transform_tracks(tracking_info: {}, pose: [float]):
    if tracking_info and len(tracking_info["objects"]) > 0:
        positions = np.array(
            [[float(obj["x"]), float(obj["y"]), float(obj["z"]), 1] for obj in tracking_info["objects"]]).T
        positions = transform_positions(positions, pose).T

        for obj, pos in zip(tracking_info["objects"], positions):
            obj["x"] = pos[0]
            obj["y"] = pos[1]
            obj["z"] = pos[2]

    return tracking_info


def _endpoint_thread(endpoint_func: Callable,
                     clients: [socket.socket],
                     semaphore: Semaphore,
                     pose: [float],
                     tracker_args: {}):
    while True:
        # update tracking info
        tracking_info = _transform_tracks(tracking_info=endpoint_func(tracker_args), pose=pose)
        if tracking_info is not None:
            bson_data = bson.dumps(tracking_info)
            size = len(bson_data)
            size_data = struct.pack("!i", size)
            data = size_data + bson_data

            # update clients
            semaphore.acquire()
            to_remove = []
            for client in clients:
                try:
                    client.sendall(data)
                except Exception:
                    logger.debug("transmit failure", exc_info=True)
                    to_remove.append(client)

            for sock in to_remove:
                clients.remove(sock)
            semaphore.release()


def _run_server(server_sockets: [socket.socket], reader, pose: [float], tracker_args):
    semaphore = Semaphore()
    clients = []

    # TODO: start pose provider thread

    # start sensor thread
    tracker_thread = Thread(target=_endpoint_thread, args=(reader, clients, semaphore, pose, tracker_args))
    tracker_thread.daemon = True
    tracker_thread.start()

    # listen for connections
    while True:
        ready, _, _ = select(server_sockets, [], [], 1)
        for server_socket in ready:
            conn, addr = server_socket.accept()
            logger.debug("client connected: {}".format(addr))
            _set_keepalive(conn)
            semaphore.acquire()
            clients.append(conn)
            semaphore.release()


def _start_endpoint(name: str, interface: str, port: int, pose: [float], tracker_type: str, tracker, tracker_args):

    ips = _get_interface_ip(interface)

    logger.debug("opening server on ({},{})".format(ips, port))
    server_sockets, ips = _open_server(ips, port)

    zeroconf, infos = _advertise_server(name, "endpoint", ips, port, {"type": tracker_type})

    try:
        _run_server(server_sockets=server_sockets,
                    reader=tracker,
                    pose=pose,
                    tracker_args=tracker_args)
    except Exception:
        logging.error("Exception while running server", exc_info=True)
    finally:
        for info in infos:
            zeroconf.unregister_service(info)
        zeroconf.close()


def start_endpoint(ctx, tracker_type: str, tracker, tracker_args=None):
    _start_endpoint(ctx.obj["name"],
                    ctx.obj["interface"],
                    ctx.obj["port"],
                    ctx.obj["pose"],
                    tracker_type,
                    tracker,
                    tracker_args)


random_default_name = uuid.uuid1()


@click.group()
@click.option('--name',
              required=False,
              default=str(random_default_name),
              help='The service name')
@click.option('--interface',
              required=False,
              default=None,
              help='The interface to bind to.')
@click.option('--port',
              required=False,
              default=8300,
              type=int,
              help='The port to bind to.')
@click.option('--pose',
              required=False,
              default=None,
              type=float,
              nargs=6,
              help='The pose of the sensor. Expressed in x y z yaw(Rx) pitch(Ry) roll(Rz) order.\n'
                   'Units are meters and degrees.\n'
                   ' +Z is the direction of sensor view. X & Y follow the right hand rule.\n'
                   'If a pose provider is installed, this will override it.')
@click.option('--pose-provider',
              required=False,
              default=None,
              type=click.Choice(list(pose_providers.keys()), case_sensitive=False),
              help='The pose provider to use for automatically determining the sensor pose.\n'
                   'NOTE: this is currently unsupported')
@click.pass_context
def cli(ctx, name: str, interface: str, port: int, pose: [float], pose_provider: str):
    global logger
    logging.basicConfig(level="DEBUG")
    logger = logging.getLogger("circum_endpoint")
    logger.debug("Loaded Plugins: {}".format(circum_sensors))

    if pose and pose_provider:
        raise Exception("--pose and --pose-provider cannot both be specified")

    if pose is None and pose_provider is None:
        pose = [0, 0, 0, 0, 0, 0]

    ctx.ensure_object(dict)
    ctx.obj["name"] = name
    ctx.obj["interface"] = interface
    ctx.obj["port"] = port
    ctx.obj["pose"] = pose
    ctx.obj["pose_provider"] = pose_provider


# register sensors
for entry_point in pkg_resources.iter_entry_points('circum.sensors'):
    circum_sensors.append(entry_point.name)
    cli.add_command(entry_point.load())


if __name__ == "__main__":
    cli(obj={})



from setuptools import setup, find_packages
from os import path


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='circum',
    version_format='{tag}',
    author="Lane Haury",
    author_email="lane@lumineerlabs.com",
    description="Circum is a set of tools for detecting and tracking moving objects" +
                "via a variety of distributed sensors.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/LumineerLabs/circum",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Operating System :: OS Independent',
    ],
    packages=find_packages(exclude=["tests"]),
    install_requires=[
        'zeroconf',
        'click',
        'numpy',
        'scipy',
        'pykalman',
        'munkres',
        'bson',
        'mock',
        'matplotlib',
        'setuptools-git-version',
    ],
    entry_points={
        'console_scripts': [
            'circum-endpoint=circum.endpoint:cli',
            'circum=circum.service:cli'
        ],
        'circum.sensors': [
            'simulator=circum.sensors.simulator:simulator',
        ],
    },
)



import numpy as np
from scipy.linalg import inv


class KalmanFilter:
    """
      A class that predicts the next state of the system given sensor measurements
      using the Kalman Filter algorithm
    """

    def __init__(self, n):
        self.n = n
        self.I = np.eye(n)  # noqa E741
        self.x = None
        self.P = None
        self.F = None
        self.Q = None

    def start(self, x: np.array, P: np.ndarray, F: np.ndarray, Q: np.ndarray):
        self.x = x
        self.P = P
        self.F = F
        self.Q = Q

    def set_Q(self, Q: np.ndarray):
        self.Q = Q

    def update_F(self, dt: float):
        self.F[0, 3], self.F[1, 4], self.F[2, 5] = dt, dt, dt

    def get_x(self):
        return self.x

    def predict(self):
        self.x = self.F @ self.x
        self.P = self.F @ self.P @ self.F.T + self.Q

    def update(self, z: np.ndarray, H: np.ndarray, Hx: np.ndarray, R: np.ndarray):
        y = z - Hx
        PHt = self.P @ H.T
        S = H @ PHt + R
        K = PHt @ inv(S)

        self.x = self.x + K @ y
        self.P = (self.I - K @ H) @ self.P
